use std::{
    env,
    fmt::Display,
    io,
    os::unix::net::UnixStream,
    path::PathBuf,
    time::{Duration, Instant},
};

use crossterm::event::{self, Event, KeyCode, KeyEvent, KeyEventKind};
use greetd::LoginResult;
use ratatui::{
    layout::{Constraint, Direction, Layout, Margin, Rect},
    style::Color,
    text::Line,
    widgets::{Block, Clear},
    DefaultTerminal, Frame,
};
use tui_rain::Rain;

mod gather;
mod greetd;

fn main() -> io::Result<()> {
    let mut terminal = ratatui::init();
    let app_result = App::default().run(&mut terminal);
    ratatui::restore();
    app_result
}
struct App {
    start_time: Instant,
    exit: bool,
    error: bool,
    focus: Focus,
    status: Status,
    user_input: SelectorInput<User>,
    session_input: SelectorInput<Session>,
    password_input: TextInput,
}

#[derive(PartialEq, Eq)]
enum Focus {
    User,
    Session,
    Password,
}

#[derive(PartialEq, Eq)]
enum Status {
    Nothing,
    Error,
    Success,
}

struct SelectorInput<T: Display> {
    values: Vec<T>,
    state: usize,
}

impl<T: Display> SelectorInput<T> {
    fn value(&self) -> &T {
        &self.values[self.state]
    }

    fn handle_key_event(&mut self, event: KeyEvent) {
        match event.code {
            KeyCode::Left => {
                self.state = (self.state + self.values.len() - 1) % self.values.len();
            }
            KeyCode::Right => {
                self.state = (self.state + 1) % self.values.len();
            }
            _ => {}
        }
    }
}

struct TextInput {
    state: String,
}

impl TextInput {
    fn value(&self) -> &str {
        &self.state
    }

    fn handle_key_event(&mut self, event: KeyEvent) {
        match event.code {
            KeyCode::Char(c) => self.state.push(c),
            KeyCode::Backspace => {
                self.state.pop();
            }
            _ => {}
        }
    }
}

#[derive(PartialEq, Eq, Clone)]
struct User {
    name: String,
    home: PathBuf,
    shell: PathBuf,
}

impl Display for User {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.name)
    }
}

#[derive(PartialEq, Eq, Clone)]
struct Session {
    name: String,
    init: PathBuf,
}

impl Display for Session {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.name)
    }
}

impl Default for App {
    fn default() -> Self {
        let start_time = Instant::now();
        let exit = false;
        let error = false;
        let focus = Focus::User;
        let status = Status::Nothing;
        let user_input = SelectorInput {
            values: gather::gather_users(),
            state: 0,
        };
        let session_input = SelectorInput {
            values: gather::gather_sessions(user_input.value()),
            state: 0,
        };
        let password_input = TextInput {
            state: String::new(),
        };

        Self {
            start_time,
            exit,
            error,
            focus,
            status,
            user_input,
            session_input,
            password_input,
        }
    }
}

impl App {
    pub fn run(&mut self, terminal: &mut DefaultTerminal) -> io::Result<()> {
        self.start_time = Instant::now();
        while !self.exit {
            terminal.draw(|frame| self.draw(frame))?;
            self.handle_events()?;
        }
        Ok(())
    }

    fn draw(&self, frame: &mut Frame<'_>) {
        self.draw_background(frame);
        self.draw_prompt_box(frame);
    }

    fn draw_background(&self, frame: &mut Frame<'_>) {
        let rain_widget = Rain::new_matrix(self.start_time.elapsed())
            .with_character_set(tui_rain::CharacterSet::UnicodeRange {
                start: 0x21,
                len: 94,
            })
            .with_bold_dim_effect(false)
            .with_rain_density(tui_rain::RainDensity::Relative { sparseness: 60 })
            .with_rain_speed(tui_rain::RainSpeed::Absolute { speed: 0.002 })
            .with_rain_speed_variance(10000.0)
            .with_tail_lifespan(Duration::from_millis(2500))
            .with_color(Color::Green)
            .with_head_color(Color::White)
            .with_noise_interval(Duration::from_millis(2000));

        frame.render_widget(rain_widget, frame.area());
    }

    fn draw_prompt_box(&self, frame: &mut Frame<'_>) {
        let columns = Layout::default()
            .direction(Direction::Horizontal)
            .constraints(vec![Constraint::Percentage(40), Constraint::Percentage(20)])
            .split(frame.area());

        let middle_rows = Layout::default()
            .direction(Direction::Vertical)
            .constraints(vec![Constraint::Percentage(40), Constraint::Length(9)])
            .split(columns[1]);

        let box_area = middle_rows[1];

        let box_widget = Block::bordered()
            .title(match self.status {
                Status::Nothing => "Hewwo! >_< :3",
                Status::Error => "Error :(",
                Status::Success => "Success!!!",
            })
            .style(match self.status {
                Status::Nothing => Color::Gray,
                Status::Error => Color::Red,
                Status::Success => Color::Green,
            });

        frame.render_widget(Clear, box_area);
        frame.render_widget(box_widget, box_area);

        let inner_area = box_area.inner(Margin::new(2, 2));

        let rows: Vec<Rect> = inner_area.rows().collect();

        let user_row =
            Line::from(format!("User: {}", self.user_input.value())).style(match self.focus {
                Focus::User => Color::White,
                _ => Color::Gray,
            });
        frame.render_widget(user_row, rows[0]);

        let session_row = Line::from(format!(
            "Session: {} {}",
            self.session_input.value(),
            self.session_input.state
        ))
        .style(match self.focus {
            Focus::Session => Color::White,
            _ => Color::Gray,
        });
        frame.render_widget(session_row, rows[2]);

        let password_row = Line::from(format!(
            "Password: {}",
            self.password_input
                .value()
                .chars()
                .map(|_| '*')
                .collect::<String>()
        ))
        .style(match self.focus {
            Focus::Password => Color::White,
            _ => Color::Gray,
        });
        frame.render_widget(password_row, rows[4]);
    }

    fn handle_events(&mut self) -> io::Result<()> {
        if event::poll(Duration::from_millis(16))? {
            match event::read()? {
                Event::Key(key_event) if key_event.kind == KeyEventKind::Press => {
                    self.handle_key_event(key_event)?;
                }
                _ => {}
            }
        }
        Ok(())
    }

    fn handle_key_event(&mut self, event: KeyEvent) -> io::Result<()> {
        match event.code {
            KeyCode::Esc => self.exit(),
            KeyCode::Enter if self.focus == Focus::Password => self.submit(),
            KeyCode::Enter => self.focus_cycle_forward(),
            KeyCode::Tab => self.focus_cycle_forward(),
            KeyCode::Down => self.focus_cycle_forward(),
            KeyCode::BackTab => self.focus_cycle_backward(),
            KeyCode::Up => self.focus_cycle_backward(),
            _ => self.focus_handle_key_event(event),
        }
        Ok(())
    }

    fn focus_cycle_forward(&mut self) {
        self.focus = match self.focus {
            Focus::User => Focus::Session,
            Focus::Session => Focus::Password,
            Focus::Password => Focus::User,
        }
    }

    fn focus_cycle_backward(&mut self) {
        self.focus = match self.focus {
            Focus::User => Focus::Password,
            Focus::Session => Focus::User,
            Focus::Password => Focus::Session,
        }
    }

    fn focus_handle_key_event(&mut self, event: KeyEvent) {
        match self.focus {
            Focus::User => {
                let old_value = self.user_input.value().clone();
                self.user_input.handle_key_event(event);
                if self.user_input.value() != &old_value {
                    self.session_input = SelectorInput {
                        values: gather::gather_sessions(self.user_input.value()),
                        state: 0,
                    }
                };
            }
            Focus::Session => self.session_input.handle_key_event(event),
            Focus::Password => self.password_input.handle_key_event(event),
        }
    }

    fn exit(&mut self) {
        self.exit = true;
    }

    fn submit(&mut self) {
        let password = self.password_input.value().to_owned();
        self.password_input.state.clear();
        if self.status == Status::Success {
            return;
        }
        match greetd::login(
            self.user_input.value(),
            self.session_input.value(),
            &password,
        )
        .unwrap()
        {
            LoginResult::Success => {
                self.error = false;
                self.status = Status::Success
            }
            LoginResult::Failure => {
                self.error = true;
                self.status = Status::Error
            }
        };
    }
}
